/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/
// modified 11.11.05 (TGGC)

#include <cassert>
#include <cstdio>
#include <fstream>
#include <sstream>
#include "safe.h"
#include "graphics.h"
#include "surface.h"
#include "d3dresource.h"
#include "input.h"
#include "dsoundresource.h"

#include "StringHelper.h"

using namespace std;

namespace FrameWork
{
	Graphics::Graphics(int width,
					   int height,
					   int bpp,
					   DWORD flags,
					   const string& windowTitle) : initialized(false)
	{
		WNDCLASSEX wc = {sizeof(WNDCLASSEX), CS_CLASSDC, windowProc, 0, 0,
						 GetModuleHandle(0), 0, LoadCursor(0, IDC_ARROW),
						 0, 0, "DUSMania Framework Window Class", 0};

		this->width = width;
		this->height = height;
		this->bpp = bpp;
		this->flags = flags;
		this->windowTitle = windowTitle;
		this->windowClass = wc;
		active = false;
		lost = false;

		cameraPosition = D3DXVECTOR2(0.0f, 0.0f);
		p_currentTexture = 0;

		// Fensterklasse registrieren
		BOOL res = RegisterClassEx(&windowClass);
		assert(res);

		// Fenstergre anpassen
		DWORD style = ((flags & GF_WINDOWED) ? WS_OVERLAPPEDWINDOW : WS_POPUP) | WS_VISIBLE;
		RECT rect; SetRect(&rect, 0, 0, width, height);
		AdjustWindowRect(&rect, style, FALSE);
		windowWidth = rect.right - rect.left;
		windowHeight = rect.bottom - rect.top;

		// Fenster erzeugen
		window = CreateWindow("DUSMania Framework Window Class",
							  windowTitle.c_str(),
							  style,
							  GetSystemMetrics(SM_CXSCREEN) / 2 - windowWidth / 2,
							  GetSystemMetrics(SM_CYSCREEN) / 2 - windowHeight / 2,
							  windowWidth,
							  windowHeight,
							  0,
							  0,
							  GetModuleHandle(0),
							  0);
		assert(window);
		active = true;

		// Direct3D initialisieren
		p_d3d = Direct3DCreate9(D3D_SDK_VERSION);
		assert(p_d3d);

		D3DDEVTYPE deviceType = (flags & GF_REFRAST) ? D3DDEVTYPE_REF : D3DDEVTYPE_HAL;

		// Back-Buffer format bestimmen
		D3DFORMAT bbFormat = D3DFMT_X8R8G8B8;
		if(flags & GF_WINDOWED)
		{
			D3DDISPLAYMODE mode;
			p_d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &mode);
			bbFormat = mode.Format;
		}
		else
		{
			bbFormat = findBestBBFormat(deviceType, bpp);
			assert(bbFormat != D3DFMT_UNKNOWN);
		}

		ZeroMemory(&d3dpp, sizeof(d3dpp));
		d3dpp.BackBufferWidth = width;
		d3dpp.BackBufferHeight = height;
		d3dpp.BackBufferFormat = bbFormat;
		d3dpp.BackBufferCount = 1;
		d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
		d3dpp.MultiSampleQuality = 0;
		d3dpp.SwapEffect = D3DSWAPEFFECT_COPY;
		d3dpp.hDeviceWindow = window;
		d3dpp.Windowed = flags & GF_WINDOWED;
		d3dpp.EnableAutoDepthStencil = FALSE;
		d3dpp.Flags = 0;
		d3dpp.FullScreen_RefreshRateInHz = (flags & GF_WINDOWED) ? 0 : findBestRefreshRate(width, height, bbFormat, 100);
		d3dpp.PresentationInterval = (flags & GF_VSYNC) ? D3DPRESENT_INTERVAL_ONE : D3DPRESENT_INTERVAL_IMMEDIATE;

		p_d3dDevice= NULL;

		// Gerteschnittstelle erzeugen
		HRESULT hres = p_d3d->CreateDevice(D3DADAPTER_DEFAULT, deviceType, window,
										   D3DCREATE_HARDWARE_VERTEXPROCESSING,
										   &d3dpp, &p_d3dDevice);
		if(FAILED(hres))
		{
			hres = p_d3d->CreateDevice(D3DADAPTER_DEFAULT, deviceType, window,
									   D3DCREATE_SOFTWARE_VERTEXPROCESSING,
									   &d3dpp, &p_d3dDevice);
		}

		assert(SUCCEEDED(hres));
		assert(p_d3dDevice);

		// Back-Buffer erstellen
		p_backBuffer = new Surface(width, height);
		assert(p_backBuffer);

		setDefaults();
		clearBackBuffer();

		initialized = true;
	}

	Graphics::~Graphics()
	{
		// Direct3D herunterfahren
		safeDelete(p_backBuffer);
		safeRelease(p_d3dDevice);
		safeRelease(p_d3d);


		DestroyWindow(window);

		// Fensterklasse lschen
		BOOL res =UnregisterClass("DUSMania Framework Window Class", GetModuleHandle(0));
	}

	bool Graphics::processWindowMessages()
	{
		MSG msg;

		// alle anstehenden Nachrichten verarbeiten
		ZeroMemory(&msg, sizeof(MSG));
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			if(msg.message == WM_QUIT) return false;
		}

		return true;
	}

	void Graphics::tryToRestore()
	{
		HRESULT hres = p_d3dDevice->TestCooperativeLevel();
		if(hres == D3DERR_DEVICELOST)
		{
			// alle Direct3D-Ressourcen freigeben
			D3DResource::handleLostDevice();
		}
		else if(hres == D3DERR_DEVICENOTRESET)
		{
			D3DResource::handleLostDevice();
			hres = p_d3dDevice->Reset(&d3dpp);
			if(SUCCEEDED(hres))
			{
				// alle Direct3D-Resourcen wiederherstellen
				D3DResource::handleResetDevice();

				setDefaults();
				clearBackBuffer();
				p_currentTexture = 0;
				p_d3dDevice->SetTexture(0, 0);
				lost = false;
			}
		}
		else if(SUCCEEDED(hres)) lost = false;
	}

	void Graphics::showBackBuffer(float alpha)
	{
		if(!active) return;

		struct Vertex
		{
			float x, y, z, rhw;
			DWORD color;
			float u, v;
		};

		const float w = static_cast<float>(width);
		const float h = static_cast<float>(height);
		const float shiftW = 0.5f / w;
		const float shiftH = 0.5f / h;
		const DWORD color = D3DXCOLOR(1.0f, 1.0f, 1.0f, alpha);
		const Vertex v[4] = {0.0f, h,		0.0f, 1.0f,		color, 0.0f + shiftW, 1.0f - shiftH,
							 0.0f, 0.0f,	0.0f, 1.0f,		color, 0.0f + shiftW, 0.0f + shiftH,
							 w, h,			0.0f, 1.0f,		color, 1.0f - shiftW, 1.0f - shiftH,
							 w, 0.0f,		0.0f, 1.0f,		color, 1.0f - shiftW, 0.0f + shiftH};

		// Rendern auf die aktuelle Oberflche beenden
		if(Surface::getCurrentRT()) Surface::getCurrentRT()->endRender();

		if(alpha == 1.0f) p_d3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		else p_d3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);

		setTexture(p_backBuffer->getTexture());
		p_d3dDevice->BeginScene();
		HRESULT hres = p_d3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, v, sizeof(v[0]));
		assert(SUCCEEDED(hres));
		p_d3dDevice->EndScene();

		if(alpha == 1.0f) p_d3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		else p_d3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);

		hres = p_d3dDevice->Present(0, 0, 0, 0);
		assert(SUCCEEDED(hres));
	}

	void Graphics::saveScreenshot(const string& filenameBegin,
								  const string& fileFormat)
	{
		// freien Dateinamen suchen
		int n = 1;
		//char filename[256];
		while(true)
		{
			std::ostringstream ossName;
			/*sprintf(filename, "%s%04d.%s", filenameBegin.c_str(), n, fileFormat.c_str());
			FILE* p_file = fopen(filename, "r");
			if(!p_file) break;
			else fclose(p_file);*/
			
			ossName << filenameBegin << n << "." << fileFormat;

			std::ifstream ifsTest( ossName.str().c_str() );

			if( !ifsTest.is_open() )
			{
				break;
			}

			n++;
		}

		std::string Format= fileFormat;

		StringHelper::toUpper( Format );

		// Bildformat festlegen
		D3DXIMAGE_FILEFORMAT imageFormat;
		if( Format == "BMP" )
		{
			imageFormat = D3DXIFF_BMP;
		}
		else if( Format ==  "TGA" )
		{
			imageFormat = D3DXIFF_TGA;
		}
		else if( Format ==  "PNG" ) 
		{
			imageFormat = D3DXIFF_PNG;
		}
		else if( Format == "JPG" )
		{
			imageFormat = D3DXIFF_JPG;
		}
		else
		{
			assert(false);
		}

		// Inhalt des echten Back-Buffers speichern
		IDirect3DSurface9* p_backBuffer;
		p_d3dDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &p_backBuffer);

		std::ostringstream ossName;
		ossName << filenameBegin << n << "." << fileFormat;

		HRESULT hres = D3DXSaveSurfaceToFile(ossName.str().c_str(), imageFormat, p_backBuffer, 0, 0);

		assert(SUCCEEDED(hres));
		p_backBuffer->Release();
	}

	void Graphics::setDefaults()
	{
		// Standardeinstellungen setzen
		p_d3dDevice->SetFVF(D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1);
		p_d3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
		p_d3dDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		p_d3dDevice->SetRenderState(D3DRS_DITHERENABLE, TRUE);
		p_d3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		p_d3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
		p_d3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
		p_d3dDevice->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE);
		p_d3dDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
		p_d3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		p_d3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
		p_d3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		p_d3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
		p_d3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
		p_d3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
		p_d3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
		p_d3dDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
	}

	void Graphics::clearBackBuffer()
	{
		// Back-Buffer leeren
		if(Surface::getCurrentRT()) Surface::getCurrentRT()->endRender();
		p_d3dDevice->BeginScene();
		p_d3dDevice->Clear(0, 0, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 0, 0);
		p_d3dDevice->EndScene();
	}

	D3DFORMAT Graphics::findBestBBFormat(D3DDEVTYPE devType,
										 int bpp)
	{
		D3DFORMAT displayFormat;
		vector<D3DFORMAT> formatList;

		switch(bpp)
		{
		case 16:
			displayFormat = D3DFMT_R5G6B5;
			formatList.push_back(D3DFMT_R5G6B5);
			formatList.push_back(D3DFMT_X1R5G5B5);
			formatList.push_back(D3DFMT_A1R5G5B5);
			break;

		case 24:
			displayFormat = D3DFMT_X8R8G8B8;
			formatList.push_back(D3DFMT_R8G8B8);
			break;

		case 32:
			displayFormat = D3DFMT_X8R8G8B8;
			formatList.push_back(D3DFMT_A2R10G10B10);
			formatList.push_back(D3DFMT_X8R8G8B8);
			formatList.push_back(D3DFMT_A8R8G8B8);
			break;
		}

		for(vector<D3DFORMAT>::const_iterator i = formatList.begin(); i != formatList.end(); i++)
		{
			// Funktioniert der Adapter mit diesem Format?
			HRESULT hres = p_d3d->CheckDeviceType(D3DADAPTER_DEFAULT, devType, displayFormat, *i, FALSE);
			if(SUCCEEDED(hres)) return *i;
		}

		return D3DFMT_UNKNOWN;
	}

	unsigned int Graphics::findBestRefreshRate(int width,
											   int height,
											   D3DFORMAT bbFormat,
											   unsigned int maxRefreshRate)
	{
		unsigned int best = 0;
		int n = p_d3d->GetAdapterModeCount(D3DADAPTER_DEFAULT, bbFormat);
		for(int i = 0; i < n; i++)
		{
			D3DDISPLAYMODE mode;
			p_d3d->EnumAdapterModes(D3DADAPTER_DEFAULT, bbFormat, i, &mode);
			if(mode.Width == width && mode.Height == height)
			{
				if(mode.RefreshRate > best &&
				   mode.RefreshRate <= maxRefreshRate) best = mode.RefreshRate;
			}
		}

		return best;
	}

	void Graphics::setTexture(IDirect3DTexture9* p_texture)
	{
		if(p_currentTexture == p_texture) return;
		p_d3dDevice->SetTexture(0, p_texture);
		p_currentTexture = p_texture;
	}

	LRESULT WINAPI windowProc(HWND window,
							  unsigned int msg,
							  WPARAM wParam,
							  LPARAM lParam)
	{
		switch(msg)
		{
		case WM_CREATE:
			break;

		case WM_CLOSE:
			PostQuitMessage(0);
			break;

		case WM_ACTIVATE:
			if(Input::isInitialized())
			{
				Graphics& graphics = Graphics::inst();
				if(LOWORD(wParam) != WA_INACTIVE)
				{
					Input::inst().acquireDevices();
					if(!graphics.active) DSoundResource::unfreeze();
					graphics.active = true;
				}
				else
				{
					Input::inst().unacquireDevices();
					if(graphics.active) DSoundResource::freeze();
					graphics.active = false;
					graphics.lost = true;
				}
			}
			break;

		case WM_PAINT:
			{
				PAINTSTRUCT ps;
				HDC dc = BeginPaint(window, &ps);
				FillRect(dc, &ps.rcPaint, static_cast<HBRUSH>(GetStockObject(BLACK_BRUSH)));
				EndPaint(window, &ps);

				if(Graphics::inst().initialized) Graphics::inst().getD3DDevice()->Present(0, 0, 0, 0);
			}
			break;

		default:
		return DefWindowProc(window, msg, wParam, lParam);
		}

		return TRUE;
	}
}
