/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/

#include <cassert>
#include "safe.h"
#include "soundinstance.h"
#include "audio.h"
#include "utils.h"
using namespace std;

namespace FrameWork
{
	SoundInstance::SoundInstance(IDirectSoundBuffer* p_sound,
								 bool persist)
	{
		assert(Audio::isInitialized());

		HRESULT hres = Audio::inst().getDSound()->DuplicateSoundBuffer(p_sound, &p_buffer);
		assert(SUCCEEDED(hres));
		assert(p_buffer);

		frozen = 0;
		virgin = true;
		this->persist = persist;
		paused = false;
		loop = false;
		volume = 0;
		pan = 0;
		frequency = 0;
	}

	SoundInstance::~SoundInstance()
	{
		safeRelease(p_buffer);
	}

	void SoundInstance::onFreeze()
	{
		if(!frozen)
		{
			DWORD status;
			p_buffer->GetStatus(&status);
			wasPlaying = status & DSBSTATUS_PLAYING;
			p_buffer->Stop();
		}

		frozen++;
	}

	void SoundInstance::onUnfreeze()
	{
		frozen--;
		if(!frozen && wasPlaying)
		{
			HRESULT hres = p_buffer->Play(0, 0, loop ? DSBPLAY_LOOPING : 0);
			assert(SUCCEEDED(hres));
		}
	}

	void SoundInstance::play(bool loop)
	{
		HRESULT hres = p_buffer->Play(0, 0, loop ? DSBPLAY_LOOPING : 0);
		assert(SUCCEEDED(hres));
		virgin = false;
		paused = false;
		this->loop = loop;
	}

	void SoundInstance::stop()
	{
		p_buffer->Stop();
		p_buffer->SetCurrentPosition(0);
	}

	void SoundInstance::pause()
	{
		p_buffer->Stop();
		paused = true;
	}

	void SoundInstance::resume()
	{
		HRESULT hres = p_buffer->Play(0, 0, loop ? DSBPLAY_LOOPING : 0);
		assert(SUCCEEDED(hres));
		virgin = false;
		paused = false;
	}

	void SoundInstance::setParameters(DWORD flags,
									  int volume,
									  int pan,
									  DWORD frequency)
	{
		if(flags & SPF_VOLUME)
		{
			HRESULT hres = p_buffer->SetVolume(volume);
			assert(SUCCEEDED(hres));
			this->volume = volume;
		}

		if(flags & SPF_PAN)
		{
			HRESULT hres = p_buffer->SetPan(pan);
			assert(SUCCEEDED(hres));
			this->pan = pan;
		}

		if(flags & SPF_FREQUENCY)
		{
			HRESULT hres = p_buffer->SetFrequency(frequency);
			assert(SUCCEEDED(hres));
			this->frequency = frequency;
		}
	}

	void SoundInstance::getParameters(DWORD flags,
									  int* p_volume,
									  int* p_pan,
									  DWORD* p_frequency) const
	{
		if(flags & SPF_VOLUME) *p_volume = volume;
		if(flags & SPF_PAN) *p_pan = pan;
		if(flags & SPF_FREQUENCY) *p_frequency = frequency;
	}

	bool SoundInstance::toBeDeleted()
	{
		if(!virgin && !persist)
		{
			DWORD status = 0;
			p_buffer->GetStatus(&status);
			if(!(status & DSBSTATUS_PLAYING) && !paused) return true;
		}

		return false;
	}
}
