/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/
// modified 12.11.05 (TGGC)

#include <cassert>
#include <cmath>
#include "safe.h"
#include "surface.h"
#include "graphics.h"
#include "utils.h"
using namespace std;

namespace FrameWork
{
	Surface* Surface::p_currentRT = 0;

	Surface::Surface(int width,
					 int height) : p_copy(0)
	{
		assert(Graphics::isInitialized());
		Graphics& graphics = Graphics::inst();

		p_texture= NULL;
		// Textur mit den gewnschten Dimensionen erstellen
		HRESULT hres = D3DXCreateTexture(graphics.getD3DDevice(), width, height, 1, D3DUSAGE_RENDERTARGET, graphics.getD3DPP().BackBufferFormat, D3DPOOL_DEFAULT, &p_texture);
		assert(SUCCEEDED(hres));
		assert(p_texture);

		// Oberflche und Oberflchenbeschreibung abfragen
		p_texture->GetSurfaceLevel(0, &p_surface);
		assert(p_surface);
		p_surface->GetDesc(&d3dsd);

		// D3DX-Schnittstelle erzeugen
		hres = D3DXCreateRenderToSurface(graphics.getD3DDevice(), width, height, graphics.getD3DPP().BackBufferFormat, FALSE, D3DFMT_UNKNOWN, &p_renderTarget);
		assert(SUCCEEDED(hres));
		assert(p_renderTarget);

		surfaceSize = D3DXVECTOR2(static_cast<float>(d3dsd.Width), static_cast<float>(d3dsd.Height));
		clear();
	}

	Surface::Surface(const string& filename,
					 bool useColorKey,
					 DWORD colorKey) : p_copy(0)
	{
		Graphics& graphics = Graphics::inst();

		// Textur laden und als Render-Target spezifizieren
		HRESULT hres = D3DXCreateTextureFromFileEx(graphics.getD3DDevice(), filename.c_str(), D3DX_DEFAULT_NONPOW2, D3DX_DEFAULT_NONPOW2, 1, D3DUSAGE_RENDERTARGET, D3DFMT_UNKNOWN, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, useColorKey ? colorKey : 0, 0, 0, &p_texture);
		assert(SUCCEEDED(hres));
		assert(p_texture);

		// Oberflche und Oberflchenbeschreibung abfragen
		p_texture->GetSurfaceLevel(0, &p_surface);
		assert(p_surface);
		p_surface->GetDesc(&d3dsd);

		// D3DX-Schnittstelle erzeugen
		hres = D3DXCreateRenderToSurface(graphics.getD3DDevice(), d3dsd.Width, d3dsd.Height, d3dsd.Format, FALSE, D3DFMT_UNKNOWN, &p_renderTarget);
		assert(SUCCEEDED(hres));
		assert(p_renderTarget);

		surfaceSize = D3DXVECTOR2(static_cast<float>(d3dsd.Width), static_cast<float>(d3dsd.Height));
		save();
	}

	Surface::Surface(HMODULE module,
					 LPCTSTR p_resource,
					 bool useColorKey,
					 DWORD colorKey) : p_copy(0)
	{
		assert(p_resource);

		Graphics& graphics = Graphics::inst();

		// Textur laden und als Render-Target spezifizieren
		HRESULT hres = D3DXCreateTextureFromResourceEx(graphics.getD3DDevice(), module, p_resource, 0, 0, 1, D3DUSAGE_RENDERTARGET, D3DFMT_UNKNOWN, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, useColorKey ? colorKey : 0, 0, 0, &p_texture);
		assert(SUCCEEDED(hres));
		assert(p_texture);

		// Oberflche und Oberflchenbeschreibung abfragen
		p_texture->GetSurfaceLevel(0, &p_surface);
		assert(p_surface);
		p_surface->GetDesc(&d3dsd);

		// D3DX-Schnittstelle erzeugen
		hres = D3DXCreateRenderToSurface(graphics.getD3DDevice(), d3dsd.Width, d3dsd.Height, d3dsd.Format, FALSE, D3DFMT_UNKNOWN, &p_renderTarget);
		assert(SUCCEEDED(hres));
		assert(p_renderTarget);

		surfaceSize = D3DXVECTOR2(static_cast<float>(d3dsd.Width), static_cast<float>(d3dsd.Height));
		save();
	}

	Surface::~Surface()
	{
		endRender();
		safeRelease(p_texture);
		safeRelease(p_surface);
		safeRelease(p_copy);
		safeRelease(p_renderTarget);

		// eventuell noch anstehende Render-Aufgaben lschen
		while(!renderQueue.empty())
		{
			delete renderQueue.front();
			renderQueue.pop();
		}
	}

	void Surface::onLostDevice()
	{
		endRender();
		safeRelease(p_texture);
		safeRelease(p_surface);
		safeRelease(p_renderTarget);

		// eventuell noch anstehende Render-Aufgaben lschen
		while(!renderQueue.empty())
		{
			delete renderQueue.front();
			renderQueue.pop();
		}
	}

	void Surface::onResetDevice()
	{
		Graphics& graphics = Graphics::inst();

		HRESULT hres = D3DXCreateTexture(graphics.getD3DDevice(), d3dsd.Width, d3dsd.Height, 1, d3dsd.Usage, d3dsd.Format, d3dsd.Pool, &p_texture);
		assert(SUCCEEDED(hres));
		assert(p_texture);

		hres = p_texture->GetSurfaceLevel(0, &p_surface);
		assert(SUCCEEDED(hres));
		assert(p_surface);

		hres = D3DXCreateRenderToSurface(graphics.getD3DDevice(), d3dsd.Width, d3dsd.Height, d3dsd.Format, FALSE, D3DFMT_UNKNOWN, &p_renderTarget);
		assert(SUCCEEDED(hres));
		assert(p_renderTarget);

		restore();
	}

	void Surface::save()
	{
		safeRelease(p_copy);

		HRESULT hres = Graphics::inst().getD3DDevice()->CreateOffscreenPlainSurface(d3dsd.Width, d3dsd.Height, d3dsd.Format, D3DPOOL_SYSTEMMEM, &p_copy, 0);
		assert(SUCCEEDED(hres));
		assert(p_copy);

		hres = Graphics::inst().getD3DDevice()->GetRenderTargetData(p_surface, p_copy);
		assert(SUCCEEDED(hres));
	}

	void Surface::restore()
	{
		if(p_copy)
		{
			HRESULT hres = Graphics::inst().getD3DDevice()->UpdateSurface(p_copy, 0, p_surface, 0);
			assert(SUCCEEDED(hres));
		}
	}

	void Surface::beginRender()
	{
		if(this == p_currentRT) return;

		// Rendern auf die vorherige Oberflche beenden
		if(p_currentRT) p_currentRT->endRender();

		// Szene beginnen
		HRESULT hres = p_renderTarget->BeginScene(p_surface, 0);
		assert(SUCCEEDED(hres));
		p_currentRT = this;
	}

	void Surface::endRender()
	{
		if(this != p_currentRT) return;

		// alle anstehenden Render-Aufgaben ausfhren
		flushRenderQueue();

		// Szene beenden
		HRESULT hres = p_renderTarget->EndScene(D3DTEXF_NONE);
		assert(SUCCEEDED(hres));
		p_currentRT = 0;
	}

	void Surface::flushRenderQueue()
	{
		if(renderQueue.empty()) return;

		Graphics& graphics = Graphics::inst();

		while(!renderQueue.empty())
		{
			RenderTask* p_rt = renderQueue.front();
			graphics.setTexture(p_rt->p_texture ? p_rt->p_texture->getTexture() : 0);

			switch(p_rt->primitiveType)
			{
			case PT_LINES:
				{
					const int n = static_cast<int>(p_rt->vertices.size()) / 2;
					HRESULT hres = graphics.getD3DDevice()->DrawPrimitiveUP(D3DPT_LINELIST, n, &p_rt->vertices[0], sizeof(Vertex));
					assert(SUCCEEDED(hres));
				}
				break;

			case PT_TRIANGLES:
				{
					const int n = static_cast<int>(p_rt->vertices.size()) / 3;
					HRESULT hres = graphics.getD3DDevice()->DrawPrimitiveUP(D3DPT_TRIANGLELIST, n, &p_rt->vertices[0], sizeof(Vertex));
					assert(SUCCEEDED(hres));
				}
				break;
			}

			delete p_rt;
			renderQueue.pop();
		}
	}

	void Surface::addRenderTask(RenderTask* p_rt)
	{
		assert(p_rt);

		if(renderQueue.empty())
		{
			// neuen Eintrag anlegen
			renderQueue.push(p_rt);
			p_rt->vertices.reserve(256);
		}
		else
		{
			// Hat der letzte Eintrag dieselben Parameter?
			RenderTask* p_lastTask = renderQueue.back();
			if(p_lastTask->primitiveType == p_rt->primitiveType &&
			   p_lastTask->p_texture == p_rt->p_texture)
			{
				// Ja - wir hngen die Vertizes einfach an. Die originale Render-Aufgabe wird gelscht.
				p_lastTask->vertices.insert(p_lastTask->vertices.end(), p_rt->vertices.begin(), p_rt->vertices.end());
				delete p_rt;
			}
			else
			{
				// neuen Eintrag anlegen
				renderQueue.push(p_rt);
				p_rt->vertices.reserve(256);
			}
		}
	}

	void Surface::clear(DWORD color)
	{
		beginRender();
		HRESULT hres = Graphics::inst().getD3DDevice()->Clear(0, 0, D3DCLEAR_TARGET, color, 0.0f, 0);
		assert(SUCCEEDED(hres));
	}

	void Surface::drawSprite(Surface* p_sprite,
							 const D3DXVECTOR2& center,
							 const D3DXVECTOR2& size,
							 float orientation,
							 DWORD color)
	{
		assert(p_sprite);
		drawSubSprite(p_sprite, D3DXVECTOR2(0.0f, 0.0f), p_sprite->getSurfaceSize(), center, size, orientation, color);
	}

	void Surface::drawSubSprite(Surface* p_sprite,
								const D3DXVECTOR2& spriteStartOnTex,
								const D3DXVECTOR2& spriteSizeOnTex,
								const D3DXVECTOR2& center,
								const D3DXVECTOR2& size,
								float orientation,
								DWORD color)
	{
		assert(this != p_sprite);
		Graphics& graphics = Graphics::inst();

		const float sinR = sinf(-orientation);
		const float cosR = cosf(-orientation);
		const D3DXVECTOR2& sizeX = 0.5f * D3DXVECTOR2(cosR, -sinR) * size.x;
		const D3DXVECTOR2& sizeY = 0.5f * D3DXVECTOR2(sinR, cosR) * size.y;

		Vertex v0, v1, v2, v3;
		v0.pos = center - sizeX - sizeY + D3DXVECTOR2(-0.5f, -0.5f);
		v1.pos = center + sizeX - sizeY + D3DXVECTOR2(-0.5f, -0.5f);
		v2.pos = center + sizeX + sizeY + D3DXVECTOR2(-0.5f, -0.5f);
		v3.pos = center - sizeX + sizeY + D3DXVECTOR2(-0.5f, -0.5f);
		const float a = 0.5f, b = -0.5f;
		v0.tex = spriteStartOnTex + D3DXVECTOR2(a, a);
		v1.tex = D3DXVECTOR2(spriteStartOnTex.x + spriteSizeOnTex.x, spriteStartOnTex.y) + D3DXVECTOR2(b, a);
		v2.tex = spriteStartOnTex + spriteSizeOnTex + D3DXVECTOR2(b, b);
		v3.tex = D3DXVECTOR2(spriteStartOnTex.x, spriteStartOnTex.y + spriteSizeOnTex.y) + D3DXVECTOR2(a, b);
		v0.color = v1.color = v2.color = v3.color = color;

		drawQuad(p_sprite, v0, v1, v2, v3);
	}

	void Surface::drawLine(Surface* p_texture,
						   const Vertex& v0,
						   const Vertex& v1)
	{
		assert(this != p_texture);
		Graphics& graphics = Graphics::inst();

		RenderTask* p_rt = new RenderTask;
		p_rt->vertices.reserve(2);
		p_rt->primitiveType = PT_LINES;
		p_rt->p_texture = p_texture;

		Vertex n0 = v0;
		Vertex n1 = v1;
		n0.pos -= graphics.getCameraPosition();
		n1.pos -= graphics.getCameraPosition();
		n0.z = n1.z = 0.0f;
		n0.rhw = n1.rhw = 1.0f;

		if(p_texture)
		{
			// Texturkoordinaten anpassen
			n0.tex.x /= p_texture->getSurfaceSize().x;
			n0.tex.y /= p_texture->getSurfaceSize().y;
			n1.tex.x /= p_texture->getSurfaceSize().x;
			n1.tex.y /= p_texture->getSurfaceSize().y;
		}

		p_rt->vertices.push_back(n0);
		p_rt->vertices.push_back(n1);

		beginRender();
		addRenderTask(p_rt);
	}

	void Surface::drawLine(const D3DXVECTOR2& p0,
						   const D3DXVECTOR2& p1,
						   DWORD color)
	{
		Vertex v0, v1;
		v0.pos = p0;
		v1.pos = p1;
		v0.color = v1.color = color;
		drawLine(0, v0, v1);
	}

	void Surface::drawTriangle(Surface* p_texture,
							   const Vertex& v0,
							   const Vertex& v1,
							   const Vertex& v2)
	{
		assert(this != p_texture);
		Graphics& graphics = Graphics::inst();

		RenderTask* p_rt = new RenderTask;
		p_rt->vertices.reserve(3);
		p_rt->primitiveType = PT_TRIANGLES;
		p_rt->p_texture = p_texture;

		Vertex n0 = v0;
		Vertex n1 = v1;
		Vertex n2 = v2;
		n0.pos -= graphics.getCameraPosition();
		n1.pos -= graphics.getCameraPosition();
		n2.pos -= graphics.getCameraPosition();
		n0.z = n1.z = n2.z = 0.0f;
		n0.rhw = n1.rhw = n2.rhw = 1.0f;

		if(p_texture)
		{
			// Texturkoordinaten anpassen
			n0.tex.x /= p_texture->getSurfaceSize().x;
			n0.tex.y /= p_texture->getSurfaceSize().y;
			n1.tex.x /= p_texture->getSurfaceSize().x;
			n1.tex.y /= p_texture->getSurfaceSize().y;
			n2.tex.x /= p_texture->getSurfaceSize().x;
			n2.tex.y /= p_texture->getSurfaceSize().y;
		}

		p_rt->vertices.push_back(n0);
		p_rt->vertices.push_back(n1);
		p_rt->vertices.push_back(n2);

		beginRender();
		addRenderTask(p_rt);
	}

	void Surface::drawQuad(Surface* p_texture,
						   const Vertex& v0,
						   const Vertex& v1,
						   const Vertex& v2,
						   const Vertex& v3)
	{
		assert(this != p_texture);
		Graphics& graphics = Graphics::inst();

		RenderTask* p_rt = new RenderTask;
		p_rt->vertices.reserve(6);
		p_rt->primitiveType = PT_TRIANGLES;
		p_rt->p_texture = p_texture;

		Vertex n0 = v0;
		Vertex n1 = v1;
		Vertex n2 = v2;
		Vertex n3 = v3;
		n0.pos -= graphics.getCameraPosition();
		n1.pos -= graphics.getCameraPosition();
		n2.pos -= graphics.getCameraPosition();
		n3.pos -= graphics.getCameraPosition();
		n0.z = n1.z = n2.z = n3.z = 0.0f;
		n0.rhw = n1.rhw = n2.rhw = n3.rhw = 1.0f;

		if(p_texture)
		{
			// Texturkoordinaten anpassen
			n0.tex.x /= p_texture->getSurfaceSize().x;
			n0.tex.y /= p_texture->getSurfaceSize().y;
			n1.tex.x /= p_texture->getSurfaceSize().x;
			n1.tex.y /= p_texture->getSurfaceSize().y;
			n2.tex.x /= p_texture->getSurfaceSize().x;
			n2.tex.y /= p_texture->getSurfaceSize().y;
			n3.tex.x /= p_texture->getSurfaceSize().x;
			n3.tex.y /= p_texture->getSurfaceSize().y;
		}

		p_rt->vertices.push_back(n0);
		p_rt->vertices.push_back(n1);
		p_rt->vertices.push_back(n2);
		p_rt->vertices.push_back(n0);
		p_rt->vertices.push_back(n2);
		p_rt->vertices.push_back(n3);

		beginRender();
		addRenderTask(p_rt);
	}
}
