/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/

#ifndef _SURFACE_H
#define _SURFACE_H

#include <string>
#include <vector>
#include <queue>
#include <d3d9.h>
#include <d3dx9.h>
#include "d3dresource.h"

namespace FrameWork
{
	class Surface;

	// Vertexstruktur
	struct Vertex
	{
		D3DXVECTOR2 pos;
		float z, rhw;
		DWORD color;
		D3DXVECTOR2 tex;
	};

	// Primitiventypen
	enum PrimitiveType
	{
		PT_LINES,		// Linien
		PT_TRIANGLES	// Dreiecke
	};

	// Render-Aufgabe
	struct RenderTask
	{
		PrimitiveType primitiveType;	// Primitiventyp (Dreiecke, Linien)
		std::vector<Vertex> vertices;	// Vertizes
		Surface* p_texture;				// Textur
	};

	class Surface : public D3DResource
	{
	public:
		Surface(int width, int height);
		Surface(const std::string& filename, bool useColorKey = true, DWORD colorKey = RGB(255, 0, 255));
		Surface(HMODULE module, LPCTSTR p_resource, bool useColorKey = true, DWORD colorKey = RGB(255, 0, 255));
		~Surface();

		void onLostDevice();
		void onResetDevice();

		// liefert die aktuelle Render-Target-Oberflche
		static Surface* getCurrentRT() {return p_currentRT;}

		// sichert den Inhalt der Oberflche
		void save();

		// stellt den Inhalt der Oberflche wieder her
		void restore();

		// vorbereiten zum Rendern auf diese Oberflche
		void beginRender();

		// Rendern beenden
		void endRender();

		// fhrt alle anstehenden Render-Aufgaben aus
		void flushRenderQueue();

		// fgt eine Render-Aufgabe zur Warteschlange hinzu
		void addRenderTask(RenderTask* p_rt);

		// setzt die Oberflche auf eine bestimmte Farbe
		void clear(DWORD color = D3DCOLOR_XRGB(0, 0, 0));

		// zeichnet ein Sprite auf die Oberflche
		void drawSprite(Surface* p_sprite, const D3DXVECTOR2& center, const D3DXVECTOR2& size, float orientation = 0.0f, DWORD color = D3DCOLOR_XRGB(255, 255, 255));

		// zeichnet einen Teil eines Sprites auf die Oberflche
		void drawSubSprite(Surface* p_sprite, const D3DXVECTOR2& spriteStartOnTex, const D3DXVECTOR2& spriteSizeOnTex, const D3DXVECTOR2& center, const D3DXVECTOR2& size, float orientation = 0.0f, DWORD color = D3DCOLOR_XRGB(255, 255, 255));
	
		// zeichnet eine Linie auf die Oberflche
		void drawLine(Surface* p_texture, const Vertex& v0, const Vertex& v1);
		void drawLine(const D3DXVECTOR2& p0, const D3DXVECTOR2& p1, DWORD color);

		// zeichnet ein Dreieck auf die Oberflche
		void drawTriangle(Surface* p_texture, const Vertex& v0, const Vertex& v1, const Vertex& v2);

		// zeichnet ein Viereck auf die Oberflche
		void drawQuad(Surface* p_texture, const Vertex& v0, const Vertex& v1, const Vertex& v2, const Vertex& v3);

		IDirect3DTexture9* getTexture() {return p_texture;}
		IDirect3DSurface9* getSurface() {return p_surface;}
		ID3DXRenderToSurface* getRenderTarget() {return p_renderTarget;}
		const D3DSURFACE_DESC& getD3DSD() const {return d3dsd;}
		const D3DXVECTOR2& getSurfaceSize() const {return surfaceSize;}

	private:
		static Surface* p_currentRT;

		IDirect3DTexture9* p_texture;
		IDirect3DSurface9* p_surface;
		IDirect3DSurface9* p_copy;
		ID3DXRenderToSurface* p_renderTarget;
		D3DSURFACE_DESC d3dsd;
		D3DXVECTOR2 surfaceSize;
		std::queue<RenderTask*> renderQueue;
	};
}

#endif
